/* Drifting_Markov.cc
 *
 * Copyright (C) 2008 Laboratoire Statistique & Genome
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "Drifting_Markov.h"

//////////////////
// CONSTRUCTORS //
//////////////////

//###################################################################################

Drifting_Markov::Drifting_Markov(const Sequence & seq)
{
  //Initialization of private members
  _order= seq.tell_order();
  _alphabet_size = seq.tell_alphabet_size(); 
  _tPi=_alphabet_size * int(pow(double(_alphabet_size),_order));
  _n = seq.tell_length();
}

//###################################################################################

//Mean Pi_t

Drifting_Markov::Drifting_Markov(const SequenceSet & seqset)
{
  //Initialization of private members
  _order= seqset(0).tell_order();
  _alphabet_size = seqset(0).tell_alphabet_size(); 
  _tPi=_alphabet_size * int(pow(double(_alphabet_size),_order));
  _n = seqset(0).tell_length();
}

//###################################################################################

Drifting_Markov::Drifting_Markov()
{
}

//###################################################################################

Drifting_Markov::Drifting_Markov(Drifting_Markov & C)
{
  _alphabet_size = C._alphabet_size;
  _tPi = C._tPi;
  _order = C._order;
  _n=C._n;
}

////////////////
// DESTRUCTOR //
////////////////

//###########################################################################################

Drifting_Markov::~Drifting_Markov(){}

//###########################################################################################

//Log_likelihood
double Drifting_Markov::log_likelihood(const Sequence & seq)
{
  int nbc=_alphabet_size;
  int nbl= int(pow(double(_alphabet_size),_order));
  int i,j;
  unsigned long t; 
  double L=0; 
  int mot;
  int code;
  if(_order==0)
    {
      for(t=_order;t<_n;t++)
	{
	  code=seq.tell_code(t);
	  if(code<0)
	    //PROBLEME DES LETTRES INVALIDES ex : aappaa  l'ordre 2 fera 0.25*Pi(0)*0.25^4
	    //au lieu de 0.25*Pi(0)*0.25^3*Pi(0)
	    {
	      L=L+ log(pow(1.0/double(nbc), double(_order+1)));
	      t=t+_order;
	    }
	  else
	    {
	      mot=code; 
	      L=L+log(Pi(t,mot)); 
	    }
	} 
    }
  else
    {
      //Code of words of size _order between ud and uf
      int uf=0;
      for(i=1;i<_order+1;i++)
	{
	  uf+=int(pow(double(_alphabet_size),i));
	}    
      int ud=0;
      for(i=1;i<_order;i++)
	{
	  ud+=int(pow(double(_alphabet_size),i));
	}

      //first order letters with stationary laws
      code=seq.tell_code(_order-1);
      Markov Pi_t(_alphabet_size,_order);
      for(i=0;i<nbl;i++)
	{
	  for(j=0;j<nbc;j++)
	    {
	      Pi_t(j+nbc*i)=Pi(_order-1,i,j);
	    }
	}
      Pi_t.compute_stat_law(true);
  
      if(code-ud<0)
	{
	  //PROBLEME DES LETTRES INVALIDES
	  double plus= log(pow(1.0/double(nbc), double(_order)));
	  L=L+ plus;
	}
      else
	{
	  double plus=log(Pi_t.Mu(code-ud));
	  L=L+ plus;
	}

      //Rest of the sequence
      for(t=_order;t<_n;t++)
	{
	  code=seq.tell_code(t);
	  if(code-uf<0)
	    //PROBLEME DES LETTRES INVALIDES ex : aappaa  l'ordre 2 fera 0.25*Pi(0)*0.25^4
	    //au lieu de 0.25*Pi(0)*0.25^3*Pi(0)
	    {
	      L=L+ log(pow(1.0/double(nbc), double(_order+1)));
	      t=t+_order;
	    }
	  else
	    {
	      mot=code-uf; 
	      L=L+log(Pi(t,mot));
	    }
	}
    }
  return L;
}

//###########################################################################################

//AIC
double Drifting_Markov::AIC(const Sequence & seq)
{
  double AIC;
  double L=log_likelihood(seq);
  AIC=2*L-2*nb_param();

  return AIC;
}

//###########################################################################################

//BIC
double Drifting_Markov::BIC(const Sequence & seq)
{
  double BIC;
  double L=log_likelihood(seq);
  BIC=2*L-nb_param()*log(double(_n));

  return BIC;
}

//###########################################################################################

double Drifting_Markov::proba_stat(unsigned long t, int letter)
{
  if ((letter<0)||(letter>_alphabet_size-1))
    {
      cout<<"Error proba_stat: letter not correct"<<endl;
      return 0;
    }

  double result=0;
  if(_order==0)
    {
      int i;
      double * law= new double[_alphabet_size];
      for (i=0;i<_alphabet_size;i++)
	{
	  law[i]=Pi(t,0,i);
	}	 
      result= law[letter];
      delete[] law;
      return result;
    }
  else
    {
      int i,j;
      int nbc=_alphabet_size;
      int nbl= int(pow(double(_alphabet_size),_order));
      Markov Pi_t(_alphabet_size,_order);
      for(i=0;i<nbl;i++)
	{
	  for(j=0;j<nbc;j++)
	    {
	      Pi_t(j+nbc*i)=Pi(t,i,j);
	    }
	}
      Pi_t.compute_stat_law(true);
   
      double * law= new double[_alphabet_size];
      for (i=0;i<_alphabet_size;i++)
	{
	  law[i]=0;
	}
      for(j=0;j<nbc;j++)
	{
	  for(i=j;i<nbl;i=i+nbc)
	    {
	      law[j]=law[j]+Pi_t.Mu(i);
	    }
	}
      result= law[letter];
      delete[] law;
      return result;
    }
  return 0;
}

//###########################################################################################

void Drifting_Markov::proba_stat_evol(int num, unsigned long in, unsigned long out, char * output_file)
{
  if(VERBOSE)
    {
      cout<<"Evolution of stationary laws..."<<endl;
    }
  if(_order==0)
    {
      int i,j;
      unsigned long t;
      unsigned long n=out-in;
      //For the too long sequence
      int pas= n/10000 +1; //"+1" for sequence size smaller than 10000
      int nbc=_alphabet_size;
     
      ostringstream trace_file;
      trace_file << output_file <<".trace_stat_"<<num<<".out";
      ofstream tr;
      tr.open(trace_file.str().c_str());

      for(t=in;t<out;t=t+pas)
	{
	  double * law= new double[_alphabet_size];
	  for (i=0;i<_alphabet_size;i++)
	    {
	      law[i]=Pi(t,0,i);
	    }
	  //Print trace file
	  tr<<t<<" ";
	  for(j=0;j<nbc;j++)
	    {
	      tr<<law[j]<<" ";
	    }
	  tr<<endl;
   
	  delete[] law;

	}
       tr.close();   
       if(VERBOSE)
	 {
	   cout<<"Output file: "<< output_file<<".trace_stat_"<<num<<".out"<<endl;
	 }
    }
  else
    {
      int i,j;
      unsigned long t;
      unsigned long n=out-in;
      //For the too long sequence
      int pas= n/10000 +1; //"+1" for sequence size smaller than 10000
      int nbc=_alphabet_size;
      int nbl= int(pow(double(_alphabet_size),_order));
      Markov Pi_t(_alphabet_size,_order);
  
      ostringstream trace_file;
      trace_file << output_file <<".trace_stat_"<<num<<".out";
      ofstream tr;
      tr.open(trace_file.str().c_str());

      for(t=in;t<out;t=t+pas)
	{
	  for(i=0;i<nbl;i++)
	    {
	      for(j=0;j<nbc;j++)
		{
		  Pi_t(j+nbc*i)=Pi(t,i,j);
		}
	    }
	  Pi_t.compute_stat_law(true);

	  double * law= new double[_alphabet_size];
	  for (i=0;i<_alphabet_size;i++)
	    {
	      law[i]=0;
	    }
	  for(j=0;j<nbc;j++)
	    {
	      for(i=j;i<nbl;i=i+nbc)
		{
		  law[j]=law[j]+Pi_t.Mu(i);
		}
	    }

	  //Print trace file
	  tr<<t<<" ";
	  for(j=0;j<nbc;j++)
	    {
	      tr<<law[j]<<" ";
	    }
	  tr<<endl;
   
	  delete[] law;

	}
      tr.close();  
      if(VERBOSE)
	{
	  cout<<"Output file: "<< output_file <<".trace_stat_"<<num<<".out"<<endl;
	}
    }
}

//###########################################################################################

void Drifting_Markov::proba_stat_evol(int num, char * output_file)
{
  proba_stat_evol(num,0,_n, output_file);
}

//###########################################################################################

void Drifting_Markov::distrib_evol(int num, unsigned long in, unsigned long out, char * output_file)
{
  if(VERBOSE)
    {
      cout<<"Evolution of distributions..."<<endl;
    }
  if(_order==0)
    {
      int i,j;
      unsigned long t;
      unsigned long n=out-in;
      //For the too long sequence
      int pas= n/10000 +1; //"+1" for sequence size smaller than 10000
      int nbc=_alphabet_size;    

      ostringstream trace_file;
      trace_file << output_file <<".trace_dist_"<<num<<".out";
   
      ofstream tr;
      tr.open(trace_file.str().c_str());

      for(t=in;t<out;t=t+pas)
	{
	  double * law= new double[_alphabet_size];
	  for (i=0;i<_alphabet_size;i++)
	    {
	      law[i]=Pi(t,0,i);
	    }
	  //Print trace file
	  tr<<t<<" ";
	  for(j=0;j<nbc;j++)
	    {
	      tr<<law[j]<<" ";
	    }
	  tr<<endl;
   
	  delete[] law;

	}
       tr.close(); 
       if(VERBOSE)
	 {
	   cout<<"Output file: "<< output_file <<".trace_dist_"<<num<<".out"<<endl;
	 }
    }
  else
    {
      int i,j,k,u,v,l;
      double mu,sum;
      unsigned long t;
      unsigned long n=out-in;
      //For the too long sequence
      int pas= n/10000 +1; //"+1" for sequence size smaller than 10000
      int nbc=_alphabet_size;
      int nbl= int(pow(double(_alphabet_size),_order));
      Markov Pi_t(_alphabet_size,_order);
      
      ostringstream trace_file;
      trace_file << output_file <<".trace_dist_"<<num<<".out";
   
      ofstream tr;
      tr.open(trace_file.str().c_str());
  
      double * law= new double[_alphabet_size];
      int * tab= new int[_order];
      double ** dist= new double * [_order];
      dist[0]= new double [_order * nbc];
      for (i=1; i<_order; i++)
	{
	  dist[i] = dist[i-1]+nbc;
	}
      for(i=0;i<_order;i++)
	{
	  for(j=0;j<nbc;j++)
	    {
	      dist[i][j]=0;  
	    }
	}

      //First order distributions
      for(k=0;k<_order;k++)
	{
	  for(i=0;i<nbl;i++)
	    {
	      for(j=0;j<nbc;j++)
		{
		  Pi_t(j+nbc*i)=Pi(in+k,i,j);
		}
	    }
	  Pi_t.compute_stat_law(true);
 	  for (i=0;i<_alphabet_size;i++)
	    {
	      law[i]=0;
	    }
	  for(j=0;j<nbc;j++)
	    {
	      for(i=j;i<nbl;i=i+nbc)
		{
		  law[j]=law[j]+Pi_t.Mu(i);
		}
	    }
	  sum=0;
 	  for (i=0;i<nbc;i++)
	    {
	      dist[k][i]=law[i];
	      dist[k][i]=abs(dist[k][i]);
	      sum+=dist[k][i];
	    }
	  //Normalization
	  for(i=0;i<nbc;i++)
	    {
	      dist[k][i]=dist[k][i]/sum;
	    }

	  //Print trace file 
	  tr<<k<<" ";
	  for(j=0;j<nbc;j++)
	    {
	      tr<<dist[k][j]<<" ";
	    }
	  tr<<endl;
	}
       //Distributions
       for(t=in+_order;t<out;t++)
 	{
	  sum=0;
	  for(v=0;v<nbc;v++)
 	    {
 	      law[v]=0;
	      for(u=0;u<nbl;u++)
		{
		  decompose(u,_order,_alphabet_size,tab);
		  mu=1;
		  for(l=0;l<_order;l++)
		    {
		      mu*=dist[l][tab[l]];
		    }
		  law[v]+=mu*Pi(t,u,v);
 		}
 	      for(l=0;l<_order-1;l++)
 		{
 		  dist[l][v]=dist[l+1][v];
 		}
 	      dist[_order-1][v]=law[v];
	      dist[_order-1][v]=abs(dist[_order-1][v]);
	      sum+=dist[_order-1][v];
 	    }
	  
	  //Normalization
 	  for(i=0;i<nbc;i++)
	    {
	      dist[_order-1][i]=dist[_order-1][i]/sum;
	    }
 	  //Print trace file 
	  if((t%pas)==0)
 	    {
 	      tr<<t<<" ";
 	      for(j=0;j<nbc;j++)
 		{
 		  tr<<dist[_order-1][j]<<" ";
 		}
 	      tr<<endl;
 	    }
	}
      tr.close();  
      if(VERBOSE)
	{
	  cout<<"Output file: "<< output_file <<".trace_dist_"<<num<<".out"<<endl;
	}

      delete[] law;
      delete[] dist;
     delete[] tab;
    }
}

//###########################################################################################

void Drifting_Markov::distrib_evol(int num, char * output_file)
{
  distrib_evol(num,0,_n, output_file);
}

//###########################################################################################

void Drifting_Markov::simulation(char * alphabet_file, ofstream& sim)
{
  Translator alpha(alphabet_file);
  int i,j;
  unsigned long t;
  int nbc=_alphabet_size;
  int nbl= int(pow(double(_alphabet_size),_order));
  sim<<">simulation"<<endl;

  if(_order==0)
    {
      
      int prec;
      string letter;
      double * vect= new double [nbc];
      for(t=_order;t<_n;t++)
	{
	  for(i=0;i<nbc;i++)
	    {
	      vect[i]=Pi(t,0,i);
	    }
	  //letter/cod with probabilities vect
	  //Stock in prec
	  //prec is prefix cod (cod begin at 0)
	  prec=Multinomial_1_Draw(nbc, vect);
	  letter=alpha.int_to_str(prec);
	  sim<<letter;
	}
      delete[] vect;
    }
  else
    {
      Markov Pi_0(_alphabet_size,_order);
 
      for(i=0;i<nbl;i++)
	{
	  for(j=0;j<nbc;j++)
	    {
	      Pi_0(j+nbc*i)=Pi(0,i,j);
	    }
	}
      Pi_0.compute_stat_law(true);

      int prec;
      int code;
      double * mu_0= new double [nbl];
      for(i=0;i<nbl;i++)
	{
	  mu_0[i]=Pi_0.Mu(i);
	}
      //letter/cod with probabilities mu_0 
      //Stock in prec
      code=Multinomial_1_Draw(nbl, mu_0); 
      prec=code;  
      string word="";  
      for(i=0;i<_order;i++)
	{
	  word=alpha.int_to_str(code%nbc)+word;
	  code=code/nbc;
	}
      sim<<word;

      string letter;
      double * vect= new double [nbc];
      for(t=_order;t<_n;t++)
	{
	  for(i=0;i<nbc;i++)
	    {
	      vect[i]=Pi(t,prec,i);
	    }
	  //letter/cod with probabilities vect
	  //Stock in prec
	  //prec is prefix cod (cod begin at 0)
	  prec=Multinomial_1_Draw(nbc, vect);
	  letter=alpha.int_to_str(prec);
	  sim<<letter;
	}  
      delete[] mu_0;
      delete[] vect;
    }
}

//###########################################################################################

//Sequence size
unsigned long Drifting_Markov::length()
{
  unsigned long n;
  n=_n;
  return n;
}

//###########################################################################################

//Order  
short Drifting_Markov::order()
{
  short n;
  n=_order;
  return n;
}

//###########################################################################################

//Alphabet_size
int Drifting_Markov::alphabet_size()
{
  int n;
  n=_alphabet_size;
  return n;
}

//###########################################################################################

void Drifting_Markov::print_Pit(unsigned long t, ofstream& pit)
{
  int nbc=_alphabet_size;
  int nbl= int(pow(double(_alphabet_size),_order));
  int i,j;
  double verif;
  for(i=0;i<nbl;i++)
    {
      for(j=0;j<nbc;j++)
	{
	  pit<<Pi(t,i,j)<<" "; 
	} 
      pit<<endl; 
    }
  pit<<endl;
}

//###########################################################################################

