/* Drifting_Markov.h
 *
 * Copyright (C) 2008 Laboratoire Statistique & Genome
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef DRIFTING_MARKOV_H
#define DRIFTING_MARKOV_H

#include "Tools.h"

using namespace std;

// Commentary in the shell
extern int VERBOSE;

//===========================================
// Lapack functions to solve linear systems
//===========================================

// System AX=B 
extern "C" { void dgesv_(int& n, int& nrhs, double* a, int& lda, int * ipiv, 
             double* b, int& ldb, int& info); }

// System AX=B (A symmetric positive definite)
extern "C" { void dposv_(char& uplo,int& n, int& nrhs, double* a, int& lda, 
             double* b, int& ldb, int& info); }

// Systeme AX=B  (A symmetric)
extern "C" { void dsysv_(char& uplo,int& n, int& nrhs, double* a, int& lda, int * ipiv,
			 double* b, int& ldb, double* lwork, int& work, int& info); }
//=======================================
// Lapack functions to inverse a matrix
//=======================================

// ipiv : 
extern "C" { void dgetrf_(int& m, int& n, double* a, int& lda, int * ipiv, int& info); }

// Inverse 
extern "C" { void dgetri_(int& n, double* a, int& lda, int * ipiv,
			 double* work, int& lwork, int& info); }

// ipiv :
extern "C" { void dsytrf_(char& uplo,int& n, double* a, int& lda, int * ipiv,
			  double* work, int& lwork, int& info); }

// Inverse (A symmetric) 
extern "C" { void dsytri_(char& uplo,int& n, double* a, int& lda, int * ipiv,
			double* work, int& info); }

//============================================
// Blas functions for the product of matrices
//============================================

// Product of two matrices
extern "C" { void dgemm_(char& transa, char& transb, int& m, int& n, int& k, double& alpha,
 		 double* a, int& lda, double* b, int& ldb, double& beta, double* c, int& ldc); }

// Product of a matrix by a symmetric matrix
extern "C" { void dsymm_(char& side, char& uplo,int& m,int& n, double& alpha, double* a, int& lda,
			 double* b, int& ldb, double& beta, double* c, int& ldc); }






//===========================================================================================
//===========================================================================================

class Drifting_Markov
{

//===========================================================================================

 protected:
  
  short _alphabet_size;            //Alphabet size
  int _tPi;                        //Size of transition matrix
  short _order;                    //Order of the Markov model
  unsigned long _n;                //Size of the modeled sequence

 
//===========================================================================================
  
 public:  

  virtual double Pi(int t, int code)=0;
  virtual double Pi(int t, int i, int j)=0; 
  virtual int nb_param()=0;
  virtual int config_file(ofstream& cf)=0;
  virtual int degree()=0;

  //--------------//
  // Constructors //
  //--------------//
  
  //DMM from one sequence
  Drifting_Markov(const Sequence & seq);
  
  //DMM from a group of sequences (mean Pi_t)
  Drifting_Markov(const SequenceSet & seqset);

  //DMM from a config_file (in sub-classes)
  Drifting_Markov();

  //Copy constructor
  Drifting_Markov(Drifting_Markov & C);
  

  //------------//
  // Destructor //
  //------------//

  virtual ~Drifting_Markov();  
      
  //----------//
  // Mthodes //
  //----------//

  //log-likelihood = L
  double log_likelihood(const Sequence & seq);
  
  //AIC = 2*L - 2*k
  double AIC(const Sequence & seq);
  
  //BIC = 2*L - k*log(n)
  double BIC(const Sequence & seq);

  //Stationary law of letter at position t
  double proba_stat(unsigned long t, int letter);
  
  //Stationary laws between positions in and out 
  //in trace_stat_segment_num.out, num is the number of the sequence in a SequenceSet
  void proba_stat_evol(int num, unsigned long in, unsigned long out, char * output_file);

  //Stationary laws in trace_stat_num.out, num is the number of the sequence in a SequenceSet
  void proba_stat_evol(int num, char * output_file);

  //Distribution between positions in and out 
  //in trace_dist_segment_num.out, num is the number of the sequence in a SequenceSet
  void distrib_evol(int num, unsigned long in, unsigned long out, char * output_file);
  
  //Distribution in trace_dist_num.out, num is the number of the sequence in a SequenceSet
  void distrib_evol(int num, char * output_file);

  //Simulation in file sim
  void simulation(char * alphabet_file, ofstream& sim);

  //Sequence size
  unsigned long length();

  //Order
  short order();

 //Alphabet_size
  int alphabet_size();

  //Print transition matrix Pi_t at position t in file pit
  void print_Pit(unsigned long t, ofstream& pit);

};

//===========================================================================================
//===========================================================================================

#endif //DRIFTING_MARKOV_H
